{
This file is part of TG9PROMCoder

TG9PROMCoder - A TG9/200 autoresponse PROM encoder


Copyright (C) 2022 G. Perotti, I1EPJ, i1epj@aricasale.it

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/.
}
unit ucoder;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, StdCtrls, Menus,
  ExtCtrls, MaskEdit, RTTICtrls, LCLType, Character, LazFileUtils, uman, uabout;

type

  { TForm1 }

  TForm1 = class(TForm)
    BBEL: TButton;
    BCR: TButton;
    BDEG: TButton;
    BFIGS: TButton;
    BLF: TButton;
    BLTRS: TButton;
    BWRU: TButton;
    Label1: TLabel;
    LCOUNT: TLabel;
    MainMenu1: TMainMenu;
    MDUMP: TMemo;
    MenuItem1: TMenuItem;
    MENCODE: TMenuItem;
    MSAVEIHEX: TMenuItem;
    MHDUMP: TMenuItem;
    MLOAD: TMenuItem;
    MSAVEBIN: TMenuItem;
    MEXIT: TMenuItem;
    MenuItem4: TMenuItem;
    MMAN: TMenuItem;
    MABOUT: TMenuItem;
    DSAVE: TSaveDialog;
    DOPEN: TOpenDialog;
    Panel1: TPanel;
    Separator1: TMenuItem;
    procedure BDEGClick(Sender: TObject);
    procedure BLTRSClick(Sender: TObject);
    procedure BFIGSClick(Sender: TObject);
    procedure BCRClick(Sender: TObject);
    procedure BLFClick(Sender: TObject);
    procedure BWRUClick(Sender: TObject);
    procedure BBELClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormKeyPress(Sender: TObject; var Key: char);
    procedure MABOUTClick(Sender: TObject);
    procedure MDUMPDblClick(Sender: TObject);
    procedure MSAVEIHEXClick(Sender: TObject);
    procedure MEXITClick(Sender: TObject);
    procedure MENCODEClick(Sender: TObject);
    procedure MHDUMPClick(Sender: TObject);
    procedure MLOADClick(Sender: TObject);
    procedure MMANClick(Sender: TObject);
    procedure MSAVEBINClick(Sender: TObject);
    function ValidateChar(Key: char; var Index: integer): boolean;
    procedure DisplayCursor;
    procedure RemoveCursor;
    procedure DumpBuf;
    procedure SaveHexFile(var f: text);

  private

  public

  end;


  T7603Buf = array[0..31] of byte;
  TCase = (LTRS,FIGS,NOTDEF);

const
  ProgramName = 'TG9/200 PROM Coder - ';
  sCursor: string = '_';
  //                    0        1         2         3         4         5          6
  //                    123456789012345678901234567890123456789012345678901123456789012
  ValidChars: string = ' ABCDEFGHIJKLMNOPQRSTUVWXYZ[]<^ -?:*3|%;8~().,9014''57=2/6+[]<^';

  ITA2_itCodes: array[1..62] of byte = (
                // LTRS
                $04,$18,$13,$0e,$12,$10,$16,$0b,
                $05,$0c,$1a,$1e,$09,$07,$06,$03,
                $0d,$1d,$0a,$14,$01,$1c,$0f,$19,
                $17,$15,$11,$1b,$1f,$02,$08,
                //FIGS
                $04,$18,$13,$0e,$12,$10,$16,$0b,
                $05,$0c,$1a,$1e,$09,$07,$06,$03,
                $0d,$1d,$0a,$14,$01,$1c,$0f,$19,
                $17,$15,$11,$15,$1f,$02,$08
  );

var
  Form1: TForm1;
  buf: T7603Buf;
  f: file of T7603Buf;
  CursorPos: integer;
  TTYCase: TCase;
  ChCount: integer;
  Encoded: boolean = FALSE;
  FileName: string = 'Noname';

implementation

{$R *.lfm}

{ TForm1 }

// Initializations at startup
procedure TForm1.FormCreate(Sender: TObject);
begin
  TTYCase := NOTDEF;
  ChCount := 0;
  LCOUNT.Caption := IntToStr(ChCount);
  DisplayCursor;
end;


// Auxiliary procedures & functions

// Save Hex Intel file
// Code taken from https://www.ibrtses.com/delphi/intelhex.html
// Copyright (99,2000) Ing.Büro R.Tschaggelar
// Adapted by I1EPJ

procedure TForm1.SaveHexFile(var f: text);

var adr, i: integer;
  usedbuf: integer=32;

function WriteLine(adr,len:integer; var buf:T7603Buf):string;

var i:integer;
    csum:byte;

begin
  csum:=0;
  result:=':'+inttohex(len,2)+inttohex(adr,4);
  csum:=len+lo(adr)+hi(adr);
  result := result+'00'; // Record type 00 (Data)
  for i:=0 to len-1 do begin
    result:=result+inttohex(buf[adr+i],2);
    csum:=csum+buf[adr+i];
  end;
  result:=result+inttohex(byte(-csum),2);
end;

function write_eof(startadr:integer):string;

var csum:byte;

begin
  result:=':00'+inttohex(startadr,4)+'01';
  csum:=hi(startadr)+lo(startadr)+1;
  result:=result+inttohex(byte(-csum),2);
end;

begin
  adr:=0;
  for i:=0 to ((usedbuf-1) div 16) do begin
    WriteLn(f,writeline(adr,16,buf));
    adr:=adr+16;
  end;
  if (usedbuf>adr) then
    WriteLn(f,writeline(adr,(usedbuf mod 16),buf));
  WriteLn(f,write_eof(0));
end;


// Dump PROM buffer in hex
procedure TForm1.DumpBuf;

var i: integer = 0;
    line: string;

begin
  MDUMP.Lines.Add('');
  line := '$'+IntToHex(i,2)+'  ';
  for i:=0 to 31 do begin
    line := Line+IntToHex(Buf[i],2)+' ';
    if ((i+1) mod 8)=0 then begin
       MDUMP.Lines.Add(line);
       line := '$'+IntToHex(i+1,2)+'  ';
    end;
  end;
end;


// Shows a _ cursor at the end of character string
procedure TForm1.DisplayCursor;

begin
  Label1.Caption := Label1.Caption+sCursor;
  CursorPos := Pos(sCursor,Label1.Caption);
end;

// Removes the _ cursor
procedure TForm1.RemoveCursor;

var s: string;

begin
  s := Label1.Caption;
  Delete(s,CursorPos,1);
  Label1.Caption := s;
end;


// Check if the typed character is a valid one
function TForm1.ValidateChar(Key: char; var Index: integer): boolean;

begin
  Index := Pos(key,ValidChars);
  if Index<>0 then
     exit(TRUE)
  else
    exit(FALSE);
end;


// Keyboard handling
procedure TForm1.FormKeyPress(Sender: TObject; var Key: char);

var s: string;
    p: integer = 0;
    c: char;
begin
  RemoveCursor;
  Key := ToUpper(Key);

  if (key=char(8)) and (ChCount>0) then begin
    s := Label1.Caption;
    c := s[CursorPos-1];
    Delete(s,CursorPos-1,1);
    Label1.Caption := s;
    Dec(ChCount);
    // If a LETTERS or a FIGURES is deleted and no more are present
    // at left switch to the opposite state
    if (c='[') and (ChCount>0) then if (s[CursorPos-2]<>'[') then TTYCASE:=LTRS;
    if (c=']') and (ChCount>0) then if (s[CursorPos-2]<>']') then TTYCase:=FIGS;
    // If no more characters left, case is undefined
    if ChCount=0 then TTYCase := NOTDEF;
    if ChCount<32 then LCOUNT.Color := clGreen else LCOUNT.Color := clRed;
  end;

  if ValidateChar(Key,p) and (ChCount<32) then begin
    if (p<>1) then begin // SPACE do not change case
      if (p<28) and ((TTYCase = FIGS) or (TTYCase = NOTDEF)) then begin
        if ChCount<32 then begin
          Label1.Caption := Label1.Caption+']'; // Insert a LTRS char
          TTYCase := LTRS;
          Inc(ChCount);
        end;
      end;

      if (p > 27) and ((TTYCase=LTRS) or (TTYCase = NOTDEF)) then begin
        if ChCount<32 then begin
          Label1.Caption := Label1.Caption+'['; // Insert a FIGS char
          TTYCase := FIGS;
          Inc(ChCount);
        end;
      end;
    end;

    if ChCount<32 then begin
      Label1.Caption := Label1.Caption + Key;
      Inc(ChCount);
    end;
  end;
  DisplayCursor;
  Key := char(0);
  LCOUNT.Caption := IntTostr(ChCount);
  if ChCount<32 then LCOUNT.Color := clGreen else LCOUNT.Color := clRed;
  Encoded := FALSE;
end;


// Menu handling

// File --> Load BIN file
procedure TForm1.MLOADClick(Sender: TObject);

var f: file of T7603Buf;
    i,j,start,stop: integer;
    b: byte;

begin
  RemoveCursor;
  DOPEN.FileName := FileName+'.bin';
  if DOPEN.Execute then begin
    {$I-}
    AssignFile(f,DOPEN.Filename);
    Reset(f);
    Read(f, Buf);
    CloseFile(f);
    {$I+}
    if IOResult<>0 then begin
       ShowMessage('Can''t open file');
       exit;
    end;

    FileName := ExtractFilenameOnly(DOPEN.FileName);
    Form1.Caption := ProgramName + Filename+'.bin';
    Label1.Caption := '';
    for i := 0 to 31 do begin
      b := Buf[i];
      if b=0 then break;
      // mdump.lines.add('b='+IntToHex(b,2));
      case b of
        $1f: begin
          Label1.Caption := Label1.Caption+']';
          TTYCase:=LTRS;
        end;
        $1b: begin
          Label1.Caption := Label1.Caption+'[';
          TTYCase :=FIGS;
        end;
        $02: Label1.Caption := Label1.Caption+'<';
        $08: Label1.Caption := Label1.Caption+'^';
        $12: if TTYCase=LTRS then
               Label1.Caption := Label1.Caption+'D'
             else
               Label1.Caption := Label1.Caption+'*';
        $1a: begin
          if TTYCase=LTRS then
            Label1.Caption := Label1.Caption+'J'
          else
            Label1.Caption := Label1.Caption+'~';
        end;
        otherwise begin
          if TTYCase=LTRS then begin
            start := 1;
            stop  := 31;
          end else begin
            start := 32;
            stop  := 62;
          end;
          for j := start to stop do begin
            if ITA2_itCodes[j]=b then begin
              Label1.Caption := Label1.Caption+ValidChars[j];
              //MDUMP.lines.add('Found $'+IntToHex(b,2)+'='+ValidChars[j]);
              break;
            end;
          end;
        end;
      end;
    end;
  end;
  ChCount := Length(Label1.Caption);
  LCOUNT.Caption := IntToStr(ChCount);
  DisplayCursor;
  MDUMP.Lines.Clear;
  DumpBuf;
end;


// File --> Save BIN file
procedure TForm1.MSAVEBINClick(Sender: TObject);

var f: file of T7603Buf;

begin
  if not Encoded then MENCODEClick(Sender);

  DSAVE.DefaultExt := '.bin';
  DSAVE.Filter := 'BIN Files|*.bin|All files|*.*';
  DSAVE.Filename := FileName+'.bin';

  if DSAVE.Execute then begin
    AssignFile(f,DSAVE.Filename);
    {$I-}
    Rewrite(f);
    Write(f, Buf);
    CloseFile(f);
    {$I+}
    if IOResult<>0 then begin
       ShowMessage('Can''t save file');
       exit;
    end;
    FileName := ExtractFilenameOnly(DSAVE.FileName);
    Form1.Caption := ProgramName + Filename+'.bin';
  end;
end;


// File --> Export IHEX File
procedure TForm1.MSAVEIHEXClick(Sender: TObject);

var f: text;

begin
  if not Encoded then MENCODEClick(Sender);

  DSAVE.DefaultExt := '.hex';
  DSAVE.Filter := 'HEX Files|*.hex|All files|*.*';
  DSAVE.FileName := FileName+'.hex';

  if DSAVE.Execute then begin
    AssignFile(f,DSAVE.Filename);
    {$I-}
    Rewrite(f);
    SaveHexFile(f);
    CloseFile(f);
    {$I+}
    if IOResult<>0 then begin
      ShowMessage('Can''t save file');
      exit;
    end;
  end;
end;


// File --> Encode
procedure TForm1.MENCODEClick(Sender: TObject);

var b: byte = 0;
  i,p: integer;
  s: string;

begin
  RemoveCursor;
  s := Label1.Caption;

  for i:= 0 to 31 do Buf[i]:=0;

  for i:=1 to ChCount do begin
    case s[i] of
      ']': b := $1F;
      '[': b := $1B;
      '<': b := $02;
      '^': b := $08;
      '*': b := $12;
      '~': b := $1A;
      '|': b := $16;
    otherwise
      p := Pos(s[i],ValidChars);
      if p<>0 then b := ITA2_itCodes[p];
    end;
    Buf[i-1] := b;
  end;
  if ChCount<32 then begin
     for i := ChCount+1 to 31 do Buf[i] := 0;
  end;
  DumpBuf;
  DisplayCursor;
  Encoded := TRUE;
end;


// File --> Dump
procedure TForm1.MHDUMPClick(Sender: TObject);
begin
  DumpBuf;
end;


// File --> Exit
procedure TForm1.MEXITClick(Sender: TObject);
begin
  Halt;
end;


// Help --> About
procedure TForm1.MABOUTClick(Sender: TObject);
begin
  FABOUT.Show;
end;


// Help --> Manual
procedure TForm1.MMANClick(Sender: TObject);
begin
  FMAN.Show;
end;


// Buttons et al handlers


// LTRS button
procedure TForm1.BLTRSClick(Sender: TObject);
begin
  RemoveCursor;
  if ChCount<32 then begin
    Label1.Caption := Label1.Caption+']';
    Inc(ChCount);
    LCOUNT.Caption := IntToStr(ChCount);
  end;
  TTYCase := LTRS;
  if ChCount=32 then LCOUNT.Color := clRed;
  DisplayCursor;
end;


// FIGS button
procedure TForm1.BFIGSClick(Sender: TObject);
begin
  RemoveCursor;
  if ChCount<32 then begin
    Label1.Caption := Label1.Caption+'[';
    Inc(ChCount);
    LCOUNT.Caption := IntToStr(ChCount);
  end;
  if ChCount=32 then LCOUNT.Color := clRed;
  TTYCase := FIGS;
  DisplayCursor;
end;


// CR button
procedure TForm1.BCRClick(Sender: TObject);
begin
  RemoveCursor;
  if ChCount<32 then begin
    Label1.Caption := Label1.Caption+'<';
    Inc(ChCount);
    LCOUNT.Caption := IntToStr(ChCount);
  end;
  if ChCount=32 then LCOUNT.Color := clRed;
  DisplayCursor;
end;


// LF button
procedure TForm1.BLFClick(Sender: TObject);
begin
  RemoveCursor;
  if ChCount<32 then begin
    Label1.Caption := Label1.Caption+'^';
    Inc(ChCount);
    LCOUNT.Caption := IntToStr(ChCount);
  end;
  if ChCount=32 then LCOUNT.Color := clRed;
  DisplayCursor;
end;


// WRU button
procedure TForm1.BWRUClick(Sender: TObject);
begin
  RemoveCursor;
  if ChCount<32 then begin
    if (TTYCase=LTRS) or (TTYCase=NOTDEF) then begin
      TTYCase := FIGS;
      Label1.Caption := Label1.Caption+'[';
      Inc(ChCount);
    end;
  end else LCOUNT.Color := clRed;
  if ChCount<32 then begin
    Label1.Caption := Label1.Caption+'*';
    Inc(ChCount);
    LCOUNT.Caption := IntToStr(ChCount);
  end else LCOUNT.Color := clRed;;
  if ChCount=32 then LCOUNT.Color := clRed;
  DisplayCursor;
end;


// BELL button
procedure TForm1.BBELClick(Sender: TObject);
begin
  RemoveCursor;
  if ChCount<32 then begin
    if (TTYCase=LTRS) or (TTYCase=NOTDEF) then begin
      TTYCase := FIGS;
      Label1.Caption := Label1.Caption+'[';
      Inc(ChCount);
    end;
  end else LCOUNT.Color := clRed;
  if ChCount<32 then begin
    Label1.Caption := Label1.Caption+'~';
    Inc(ChCount);
    LCOUNT.Caption := IntToStr(ChCount);
  end else LCOUNT.Color := clRed;;
  if ChCount=32 then LCOUNT.Color := clRed;
  DisplayCursor;
end;


// ° button
procedure TForm1.BDEGClick(Sender: TObject);
begin
  RemoveCursor;
  if ChCount<32 then begin
    if (TTYCase=LTRS) or (TTYCase=NOTDEF) then begin
      TTYCase := FIGS;
      Label1.Caption := Label1.Caption+'[';
      Inc(ChCount);
    end;
  end else LCOUNT.Color := clRed;
  if ChCount<32 then begin
    Label1.Caption := Label1.Caption+'|';
    Inc(ChCount);
    LCOUNT.Caption := IntToStr(ChCount);
  end else LCOUNT.Color := clRed;;
  if ChCount=32 then LCOUNT.Color := clRed;
  DisplayCursor;
end;


// Double click on dump memo
procedure TForm1.MDUMPDblClick(Sender: TObject);
begin
  MDUMP.Lines.Clear;
end;

end.

